#version 400 compatibility
#include "/lib/global.glsl"
#include "/lib/buffer.glsl"
#include "/lib/util/math.glsl"

/* ------ uniforms ------ */

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex6;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform int isEyeInWater;
uniform int worldTime;
uniform int frameCounter;

uniform float far;
uniform float near;
uniform float aspectRatio;
uniform float frameTimeCounter;
uniform float viewWidth;
uniform float viewHeight;
uniform float rainStrength;
uniform float wetness;
uniform float eyeAltitude;

uniform ivec2 eyeBrightnessSmooth;

uniform vec3 cameraPosition;

uniform vec3 skyColor;
uniform vec3 fogColor;

uniform mat4 gbufferModelView;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjection;
uniform mat4 gbufferProjectionInverse;


/* ------ inputs from vertex stage ------ */

in vec2 coord;

flat in vec3 sunVector;
flat in vec3 moonVector;
flat in vec3 lightVector;
flat in vec3 upVector;

flat in float timeSunrise;
flat in float timeNoon;
flat in float timeSunset;
flat in float timeNight;
flat in float timeMoon;
flat in float timeLightTransition;
flat in float timeSun;

flat in vec3 colSunlight;
flat in vec3 colSkylight;
flat in vec3 colSky;
flat in vec3 colHorizon;


/* ------ structs ------ */

struct sceneData {
    vec3 albedo;
    vec3 normal;
    vec2 lightmap;
    vec4 sample2;
    vec4 sample3;
} scene;

struct depthData {
    float depth;
    float linear;
    float solid;
    float solidLin;
} depth;

struct positionData {
    vec3 camera;
    vec3 view;
    vec3 world;
} pos;

struct vectorData {
    vec3 sun;
    vec3 moon;
    vec3 light;
    vec3 up;
    vec3 view;
} vec;

struct lightData {
    vec3 sun;
    vec3 sky;
} light;

vec3 returnCol  = vec3(0.0);
bool translucency = false;
float cloudAlpha    = 0.0;


/* ------ includes ------ */

#include "/lib/util/decode.glsl"
#include "/lib/util/decodeIn.glsl"
#include "/lib/util/colorConversion.glsl"
#include "/lib/util/depth.glsl"
#include "/lib/util/positions.glsl"
#include "/lib/util/dither.glsl"
#include "/lib/util/taaJitter.glsl"
#include "/lib/util/encode.glsl"

vec3 unpackNormal(vec3 x) {
    return x*2.0-1.0;
}

vec3 fogVanilla = toLinear(fogColor)*2.0*(1.0-timeNight*0.4);

#include "/lib/nature/getSky.glsl"

void simpleFog() {
    vec3 fogVanilla = toLinear(fogColor)*2.0;
    
    float falloff   = saturate(length(pos.world.xyz-pos.camera)/far);
        falloff     = linStep(falloff, s_fogStart, 0.999);
        falloff     = pow(falloff, s_fogExp);
    
    vec3 skyCol     = falloff>0.0 ? getSky(vec.view) : fogVanilla;

    returnCol       = mix(returnCol, skyCol, falloff);
}
void simpleFogEyeInWater() {
    vec3 fogVanilla = toLinear(fogColor)*2.0;

    vec3 wPosSolid  = getWorldpos(depth.solid);
    vec3 wPos       = pos.world;
    float solidDistance = length(wPosSolid-pos.camera)/(far*0.8);
    float transDistance  = length(wPos-pos.camera)/(far*0.8);
    
    float falloff   = saturate(solidDistance-transDistance);
        falloff     = linStep(falloff, 0.35, 0.999);
        falloff     = pow2(falloff);
    
    vec3 skyCol     = falloff>0.0 ? getSky(vec.view) : fogVanilla;

    returnCol       = mix(returnCol, skyCol, falloff);
}

void underwaterFog() {
    vec3 wPosSolid  = getWorldpos(depth.solid).xyz;
    vec3 wPos       = pos.world;

    float solidDistance = length(wPosSolid-pos.camera)/far16;
    float transDistance  = length(wPos-pos.camera)/far16;

    float falloff   = isEyeInWater==1 ? transDistance : solidDistance-transDistance;
        falloff     = saturate(falloff);
        falloff     = linStep(falloff, 0.0, 0.2);
        falloff     = 1.0-pow2(1.0-falloff);

    vec3 fogCol     = (colSunlight*sunlightLuma+colSkylight*0.1)*vec3(0.1, 0.4, 1.0)*0.1;

    float caveFix   = isEyeInWater==1 ? 1.0 : linStep(eyeBrightnessSmooth.y/240.0, 0.0, 0.5);
    
    returnCol       = mix(returnCol, fogCol, falloff*caveFix);
}

void applyTranslucents() {
    vec4 translucents   = texture(colortex6, coord)*vec4(vec3(20.0), 1.0);
    if (eyeAltitude > s_cloudAltitude-20) translucents.a *= 1.0-cloudAlpha;

    if (translucency) {
        vec3 translucentColor = saturate(normalize(translucents.rgb));
            translucentColor = mix(vec3(1.0), translucentColor, translucents.a);
        returnCol *= translucentColor;
    }
    
    returnCol       = mix(returnCol, translucents.rgb, translucents.a);
}

void lavaFog() {
    vec3 wPosSolid  = getWorldpos(depth.solid).xyz;
    vec3 wPos       = pos.world;

    float solidDistance = length(wPosSolid-pos.camera)/far16;

    float falloff   = solidDistance;
        falloff     = saturate(falloff);
        falloff     = linStep(falloff, 0.0, 0.14);
        falloff     = 1.0-pow2(1.0-falloff);

    vec3 fogCol     = vec3(1.0, 0.3, 0.02)*1.5;
    
    returnCol       = mix(returnCol, fogCol, falloff);
}

#ifdef s_clouds
vec4 bilinear(sampler2D tex, vec2 coord) {
    ivec2 texSize = textureSize(tex, 0)*8;
    vec2 texelSize = vec2(1.0)/texSize;
    vec4 p0q0 = texture(tex, coord);
    vec4 p1q0 = texture(tex, coord + vec2(texelSize.x, 0));

    vec4 p0q1 = texture(tex, coord + vec2(0, texelSize.y));
    vec4 p1q1 = texture(tex, coord + vec2(texelSize.x , texelSize.y));

    float a = fract(coord.x * texSize.x);

    vec4 pInterp_q0 = mix(p0q0, p1q0, a);
    vec4 pInterp_q1 = mix(p0q1, p1q1, a);

    float b = fract(coord.y*texSize.y);
    return mix(pInterp_q0, pInterp_q1, b);
}

float noise2DCloud(in vec2 coord, in vec2 offset, float size) {
    coord += offset;
    coord = coord*size;
    coord /= 1024;

    return bilinear(colortex4, coord).a;
}
float heightFade(vec3 wPos, float limit, in float smoothing) {
    float density   = sstep(wPos.y, limit-smoothing/2, limit+smoothing/2);
    return density;
}

const float cloudAltitude = s_cloudAltitude;
const float cloudDepth    = 20.0;

float cloudDensityPlane(vec3 pos) {
    const float lowEdge     = cloudAltitude-cloudDepth/2;
    const float highEdge    = cloudAltitude+cloudDepth/2;
    float size              = 0.15;
    vec2 coord              = pos.xz;
    float height            = heightFade(pos, lowEdge, 2.0)-heightFade(pos, highEdge, 2.0);

    float animTick          = frameTimeCounter*1.0;
    vec2 animVec            = vec2(animTick, 0.0);
    
    float shape;
    shape           = noise2DCloud(coord, animVec, size);

    #ifdef setCVtexErosion
    float low       = pow3(1.0-linStep(pos.y, lowEdge, cloudAltitude-cloudDepth/4));
    float high      = pow3(linStep(pos.y, cloudAltitude+cloudDepth/4, highEdge));
    shape          -= low+high;
    #endif

    return saturate(shape*2.0*height-0.01);
}
float scatterIntegral(float transmittance, const float coeff) {
    float a   = -1.0/coeff;
    return transmittance * a - a;
}

void getCloudLighting(inout float scatter, in vec3 pos, in float transmittance, in float stepTransmittance, const float lowEdge, const float highEdge) {
    float scatterInt    = scatterIntegral(stepTransmittance, 1.0);
    float s         = sstep(pos.y, lowEdge, highEdge)*0.2+pow3(linStep(pos.y, cloudAltitude, highEdge))*0.5;
    scatter += s*scatterInt*transmittance;
}

void cloudVolumetricVanilla(inout vec3 scenecol) {
    const int steps         = s_cloudSamples;
    const float lowEdge     = cloudAltitude-cloudDepth/2;
    const float highEdge    = cloudAltitude+cloudDepth/2;
    const float threshold   = 0.005;
    const float density     = 1.0;

    /* --- calculate spheres --- */
    vec3 wvec       = mat3(gbufferModelViewInverse)*vec.view;
    vec2 psphere    = rsi((planetRadius+eyeAltitude)*vec.up, vec.view, planetRadius);
    bool visible    = !((eyeAltitude<lowEdge && psphere.y>0.0) || (eyeAltitude>highEdge && wvec.y>0.0));

    if (visible) {
        vec2 bsphere    = rsi(vec3(0.0, 1.0, 0.0)*planetRadius+eyeAltitude, wvec, planetRadius+lowEdge);
        vec2 tsphere    = rsi(vec3(0.0, 1.0, 0.0)*planetRadius+eyeAltitude, wvec, planetRadius+highEdge);
    
        float startdist = eyeAltitude>highEdge ? tsphere.x : bsphere.y;
        float enddist   = eyeAltitude>highEdge ? bsphere.x : tsphere.y;

        vec3 startpos   = wvec*startdist;
        vec3 endpos     = wvec*enddist;

        float mrange    = (1.0-saturate((eyeAltitude-highEdge)*0.2)) * (1.0-saturate((lowEdge-eyeAltitude)*0.2));
            mrange      = mix(1.0, mrange, float(!mask.terrain));

        startpos        = mix(startpos, gbufferModelViewInverse[3].xyz, mrange);
        endpos          = mix(endpos, (pos.world-pos.camera)*(!mask.terrain ? (highEdge/64.0) : 1.0), mrange);

        float dither    = ditherDynamic;

        vec3 rstep      = (endpos-startpos)/steps;
        vec3 rpos       = rstep*dither + startpos + pos.camera;

        float rlength   = length(rstep);

        float scatter   = 0.0;
        float transmittance = 1.0;
        float cloud     = 0.0;
        float fade      = 1.0;

        vec3 sunlight   = mix(light.sun*2.0, light.sky*4.5+fogVanilla*0.15, timeLightTransition);
        vec3 skylight   = colSky*0.5+fogVanilla*0.05*(1.0-timeLightTransition*0.4);

        for (int i = 0; i<steps; ++i, rpos += rstep) {
            if (transmittance<threshold) break;
            float dist  = length(rpos-pos.camera);
            float dfade = linStep(dist, 500.0, 2000.0);
            if (finv(dfade)<0.01) continue;
            
            float oD        = cloudDensityPlane(rpos)*rlength*density;

            if (oD <= 0.0) continue;

            cloud      += oD;
            float stepT = exp2(-oD*1.11*invLog2);

            fade       -= dfade*transmittance;

            getCloudLighting(scatter, rpos, transmittance, stepT, lowEdge, highEdge);

            transmittance *= stepT;
        }

        vec3 color          = mix(skylight, sunlight, scatter*0.98+0.02);
            //color           = mix(color, scene.albedo, saturate((1.0-pow2(1.0-fadeAlpha))));

        cloud   = linStep(cloud, 0.04, 1.0)*pow2(saturate(fade));
        cloudAlpha          = linStep(cloud, 0.0, 0.99);
        scenecol            = mix(scenecol, color, cloud*0.999);
    }
}
#endif

void main() {
    scene.albedo    = texture(colortex0, coord).rgb;
    scene.normal    = unpackNormal(texture(colortex1, coord).rgb);
    scene.sample2   = texture(colortex2, coord);
    scene.sample3   = texture(colortex3, coord);

    decodeData();

    depth.depth     = texture(depthtex0, coord).x;
    depth.linear    = depthLin(depth.depth);
    depth.solid     = texture(depthtex1, coord).x;
    depth.solidLin  = depthLin(depth.solid);

    translucency = depth.solid>depth.depth;

    pos.camera      = cameraPosition;
    pos.view        = getViewpos(depth.depth);
    pos.world       = toWorldpos(pos.view);

    vec.sun         = sunVector;
    vec.moon        = moonVector;
    vec.light       = lightVector;
    vec.up          = upVector;
    vec.view        = normalize(pos.view);

    returnCol       = scene.albedo;

    light.sun       = colSunlight*sunlightLuma;
    light.sun       = mix(light.sun, vec3(vec3avg(light.sun))*0.15, rainStrength*0.95);
    light.sky       = colSkylight*skylightLuma;
    light.sky       = mix(light.sky, vec3(vec3avg(light.sky))*0.4, rainStrength*0.95);

    bool water = pbr.roughness < 0.006;

    #ifdef s_clouds
        cloudVolumetricVanilla(returnCol);
    #endif

    if (isEyeInWater==0 && (mask.terrain || translucency)) {
        if (water) underwaterFog();
        applyTranslucents();

        #ifdef setFog
            if (translucency) simpleFog();
        #endif
        
    } else if (isEyeInWater==1) {
        if (mask.terrain) simpleFogEyeInWater();
        applyTranslucents();
        underwaterFog();
    }

    if (isEyeInWater==2) lavaFog();

    if (scene.sample3.r>0.75 && scene.sample3.r<0.95) returnCol = mix(returnCol, vec3(vec3avg(light.sun)*0.1), 0.3);

    #ifdef s_godrays
        vec4 translucentsample = texture(colortex6, coord);

        vec3 godrayColor    = mix(normalize(translucentsample.rgb), vec3(1.0), translucentsample.a);
            if (!translucency) godrayColor = vec3(1.0);

        float godrayAlpha = 1.0-float(mask.terrain);
            godrayAlpha *= 1.0-cloudAlpha;
            godrayAlpha *= 1.0-translucentsample.a;

        vec3 return6    = vec3(godrayAlpha)*godrayColor;
    #else
        vec3 return6 = vec3(0.0);
    #endif

    /*DRAWBUFFERS:06*/
    gl_FragData[0]  = makeSceneOutput(returnCol);
    gl_FragData[1]  = vec4(return6, 1.0);
}