#version 400 compatibility
#include "/lib/global.glsl"
#include "/lib/buffer.glsl"
#include "/lib/util/math.glsl"

uniform sampler2D colortex4;

uniform int frameCounter;

uniform float frameTimeCounter;
uniform float rainStrength;
uniform float wetness;
uniform float eyeAltitude;

uniform vec3 cameraPosition;

uniform vec3 fogColor;

uniform mat4 gbufferModelView;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjection;
uniform mat4 gbufferProjectionInverse;

/* ------ inputs from vertex stage ------ */

in vec2 coord;

flat in vec3 sunVector;
flat in vec3 moonVector;
flat in vec3 lightVector;
flat in vec3 upVector;

flat in float timeSunrise;
flat in float timeNoon;
flat in float timeSunset;
flat in float timeNight;
flat in float timeMoon;
flat in float timeLightTransition;
flat in float timeSun;

flat in vec3 colSunlight;
flat in vec3 colSkylight;
flat in vec3 colSky;
flat in vec3 colHorizon;
flat in vec3 colSunglow;

/* ------ structs ------ */

struct positionData {
    vec3 camera;
} pos;

struct vectorData {
    vec3 sun;
    vec3 moon;
    vec3 light;
    vec3 up;
} vec;

struct lightData {
    vec3 sun;
    vec3 sky;
} light;

vec3 returnCol  = vec3(0.0);
bool translucency = false;
bool water = false;

#include "/lib/util/colorConversion.glsl"
#include "/lib/util/dither.glsl"

vec3 unpackNormal(vec3 x) {
    return x*2.0-1.0;
}

vec3 fogVanilla = toLinear(fogColor)*10.0*mix(normalize(colSunlight), vec3(1.0), timeNight*0.25);

vec3 getSky(in vec3 viewvec) {
    vec3 nFrag      = -viewvec;
    vec3 hVec       = normalize(-vec.up+nFrag);
    vec3 hVec2      = normalize(vec.up+nFrag);
    vec3 sgVec      = normalize(vec.sun+nFrag);
    vec3 mgVec      = normalize(vec.moon+nFrag);

    float hTop      = dot(hVec, nFrag);
    float hBottom   = dot(hVec2, nFrag);

    float horizonFade = linStep(hBottom, 0.3, 0.8);
        horizonFade = pow4(horizonFade)*0.75;

    float lowDome   = linStep(hBottom, 0.66, 0.71);
        lowDome     = pow3(lowDome);

    float horizonGrad = 1.0-max(hBottom, hTop);

    float horizon   = linStep(horizonGrad, 0.15, 0.31);
        horizon     = pow6(horizon)*0.8;

    float sunGrad   = 1.0-dot(sgVec, nFrag);
    float moonGrad  = 1.0-dot(mgVec, nFrag);

    float sunGlow   = linStep(sunGrad, 0.7, 0.98);
        sunGlow     = pow3(sunGlow)*0.4;
        sunGlow    *= 1.0-rainStrength*0.9;

    float moonGlow  = pow(moonGrad*0.85, 15.0);
        moonGlow    = saturate(moonGlow*1.05)*0.8;

    vec3 sunLight   = colSunlight*5;
    vec3 moonColor  = vec3(0.55, 0.75, 1.0)*0.1;

    vec3 sky        = mix(colSky, fogVanilla, horizonFade);
        sky         = mix(sky, fogVanilla, horizon);
        sky         = mix(sky, fogVanilla, lowDome);
        sky         = mix(sky, sunLight, sunGlow*(1.0-timeNight));
        sky         = mix(sky, moonColor, moonGlow*timeNight);

    return sky;
}

vec3 cartToSphere(vec2 coord) {
    coord *= vec2(tau, pi);
    vec2 lon = sincos(coord.x) * sin(coord.y);
    return vec3(lon.x, cos(coord.y), lon.y);
}

#ifdef s_cloudReflection
vec4 bilinear(sampler2D tex, vec2 coord) {
    ivec2 texSize = textureSize(tex, 0)*8;
    vec2 texelSize = vec2(1.0)/texSize;
    vec4 p0q0 = texture(tex, coord);
    vec4 p1q0 = texture(tex, coord + vec2(texelSize.x, 0));

    vec4 p0q1 = texture(tex, coord + vec2(0, texelSize.y));
    vec4 p1q1 = texture(tex, coord + vec2(texelSize.x , texelSize.y));

    float a = fract(coord.x * texSize.x);

    vec4 pInterp_q0 = mix(p0q0, p1q0, a);
    vec4 pInterp_q1 = mix(p0q1, p1q1, a);

    float b = fract(coord.y*texSize.y);
    return mix(pInterp_q0, pInterp_q1, b);
}

float noise2DCloud(in vec2 coord, in vec2 offset, float size) {
    coord += offset;
    coord = coord*size;
    coord /= 1024;

    return bilinear(colortex4, coord).a;
}
float heightFade(vec3 wPos, float limit, in float smoothing) {
    float density   = sstep(wPos.y, limit-smoothing/2, limit+smoothing/2);
    return density;
}

const float cloudAltitude = s_cloudAltitude;
const float cloudDepth    = 20.0;

float cloudDensityPlane(vec3 pos) {
    const float lowEdge     = cloudAltitude-cloudDepth/2;
    const float highEdge    = cloudAltitude+cloudDepth/2;
    float size              = 0.15;
    vec2 coord              = pos.xz;
    float height            = heightFade(pos, lowEdge, 2.0)-heightFade(pos, highEdge, 2.0);

    float animTick          = frameTimeCounter*1.0;
    vec2 animVec            = vec2(animTick, 0.0);
    
    float shape;
    shape           = noise2DCloud(coord, animVec, size);

    #ifdef setCVtexErosion
    float low       = pow3(1.0-linStep(pos.y, lowEdge, cloudAltitude-cloudDepth/4));
    float high      = pow3(linStep(pos.y, cloudAltitude+cloudDepth/4, highEdge));
    shape          -= low+high;
    #endif

    return saturate(shape*2.0*height-0.01);
}
float scatterIntegral(float transmittance, const float coeff) {
    float a   = -1.0/coeff;
    return transmittance * a - a;
}

void getCloudLighting(inout float scatter, in vec3 pos, in float transmittance, in float stepTransmittance, const float lowEdge, const float highEdge) {
    float scatterInt    = scatterIntegral(stepTransmittance, 1.0);
    float s         = sstep(pos.y, lowEdge, highEdge)*0.2+pow3(linStep(pos.y, cloudAltitude, highEdge))*0.5;
    scatter += s*scatterInt*transmittance;
}

void cloudVolumetricVanilla(inout vec3 scenecol, in vec3 viewvec) {
    const int steps         = 6;
    const float lowEdge     = cloudAltitude-cloudDepth/2;
    const float highEdge    = cloudAltitude+cloudDepth/2;
    const float threshold   = 0.01;
    const float density     = 1.0;

    /* --- calculate spheres --- */
    vec3 wvec       = mat3(gbufferModelViewInverse)*viewvec;
    vec2 psphere    = rsi((planetRadius+eyeAltitude)*vec.up, viewvec, planetRadius);
    bool visible    = !((eyeAltitude<lowEdge && psphere.y>0.0) || (eyeAltitude>highEdge && wvec.y>0.0));

    if (visible) {
        vec2 bsphere    = rsi(vec3(0.0, 1.0, 0.0)*planetRadius+eyeAltitude, wvec, planetRadius+lowEdge);
        vec2 tsphere    = rsi(vec3(0.0, 1.0, 0.0)*planetRadius+eyeAltitude, wvec, planetRadius+highEdge);
    
        float startdist = eyeAltitude>highEdge ? tsphere.x : bsphere.y;
        float enddist   = eyeAltitude>highEdge ? bsphere.x : tsphere.y;

        vec3 startpos   = wvec*startdist;
        vec3 endpos     = wvec*enddist;

        float dither    = ditherStatic;

        vec3 rstep      = (endpos-startpos)/steps;
        vec3 rpos       = rstep*dither + startpos + pos.camera;

        float rlength   = length(rstep);

        float scatter   = 0.0;
        float transmittance = 1.0;
        float cloud     = 0.0;
        float fade      = 1.0;

        vec3 sunlight   = mix(light.sun*2.0, light.sky*4.5+fogVanilla*0.15, timeLightTransition);
        vec3 skylight   = colSky*0.5+fogVanilla*0.05*(1.0-timeLightTransition*0.4);

        for (int i = 0; i<steps; ++i, rpos += rstep) {
            if (transmittance<threshold) break;
            float dist  = length(rpos-pos.camera);
            float dfade = linStep(dist, 500.0, 2000.0);
            if (finv(dfade)<0.01) continue;
            
            float oD        = cloudDensityPlane(rpos)*rlength*density;

            if (oD <= 0.0) continue;

            cloud      += oD;
            float stepT = exp2(-oD*1.11*invLog2);

            fade       -= dfade*transmittance;

            getCloudLighting(scatter, rpos, transmittance, stepT, lowEdge, highEdge);

            transmittance *= stepT;
        }

        vec3 color          = mix(skylight, sunlight, scatter*0.98+0.02);
            //color           = mix(color, scene.albedo, saturate((1.0-pow2(1.0-fadeAlpha))));

        cloud   = linStep(cloud, 0.04, 1.0)*pow2(saturate(fade));
        scenecol            = mix(scenecol, color, cloud*0.999);
    }
}
#endif

void main() {
    pos.camera      = cameraPosition;

    vec.sun         = sunVector;
    vec.moon        = moonVector;
    vec.light       = lightVector;
    vec.up          = upVector;

    light.sun       = colSunlight*sunlightLuma;
    light.sun       = mix(light.sun, vec3(vec3avg(light.sun))*0.15, rainStrength*0.95);
    light.sky       = mix(colSkylight, colSky*1.2, 0.66)*skylightLuma;
    light.sky       = mix(light.sky, vec3(vec3avg(light.sky))*0.4, rainStrength*0.95);

    vec3 spherevec  = cartToSphere(coord)*vec3(1.0, 1.0, -1.0);
        spherevec   = mat3(gbufferModelView)*spherevec;

    vec3 skyref     = getSky(spherevec);

    #ifdef s_cloudReflection
        cloudVolumetricVanilla(skyref, spherevec);
    #endif

    /*DRAWBUFFERS:4*/
    gl_FragData[0]  = makeSceneOutput(skyref);
}